/*************************************************************************
*  3ds2obj                                                               *
*                                                                        *  
*  converts 3DS files to blenders vidioscape files                       *
*                                                                        *  
*  Copyright (C) 1999 Patrik Adamski                                     *
*                                                                        *
*  This program is free software; you can redistribute it and/or modify  *
*  it under the terms of the GNU General Public License as published by  *
*  the Free Software Foundation; either version 2 of the License, or     *
*  (at your option) any later version.                                   *
*                                                                        *
*  This program is distributed in the hope that it will be useful,       *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*  GNU General Public License in the file COPYING for more details.      *
*                                                                        *
*  You should have received a copy of the GNU General Public License     *
*  along with this program; if not, write to the Free Software           *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* #define DEBUG 1 */

struct _face      
{
   int a;                      /* index of vertix in the object */
   int b;                      /* in 3DS files only triangles ! */ 
   int c;
   unsigned char options;      /* face options (unused) */
};

struct _object
{
   int type;                   /* 0 = unknown, 1 = mesh, 2 = lamp, 3 = spot */
   unsigned long faces;        /* number of faces */
   unsigned long vertices;     /* number of vertices */
   float hotspot;              /* hotspot */
   float falloff;              /* falloff for spot */
   char name[255];             /* object name */
   float *x;                   /* pointer to an array of vertex coordinates (x)
                                  or x of a lamp x[0] , x[1] spot */ 
   float *y;                   /* pointer to an array of vertex coordinates (y)
                                  or y of a lamp y[0] , y[1] spot */    
   float *z;                   /* pointer to an array of vertex coordinates (z)
                                  or z of a lamp z[0] , z[1] spot */   
   struct _face * face;        /* pointer to an array of faces */ 
   struct _object *next;
} *first_layer = NULL, *layer = NULL, *last_layer = NULL;

void error(char *string)
{
   printf("%s\n",string);
   exit(-1);
}

int main(int argc, char * argv[])
{
   unsigned long x,n;                                     
   FILE * fp, * fp_out;             /* filepointer for read and write */             
   unsigned short chunk_id;         /* chunk ID of 3DS file */  
   unsigned long chunk_length;      /* length of current chunk */ 
   unsigned long main_length = 1;   /* length of main chunk */ 
   char object_name[256];           /* puffer for current object name */
   char out_name[256];              /* outputname without extension */
   char out_file[256];              /* outputname with extension */ 
   char letter;                     /* puffer to bulding strings */
   float a,b,c;                     /* math stuff */    
   double A, B, C;
   /* check for arguments */
   if(argc != 3)
   {
       printf("use         : 3ds2obj  inputfile  outputname\n");
       printf("for Example : 3ds2obj  test.3ds test\n");
       printf("dont't use extensions for the outputname !\n");  
       exit(-1);     
   }

   /* open inputfile */
   if((fp = fopen(argv[1],"r"))==NULL)
   {
      error("can't open file");
   }

   /* reading outname */
   sprintf(out_name,"%s",argv[2]);

   
   /* read chunks */
   while((!feof(fp)) && (ftell(fp) < main_length))
   {
      chunk_id     = (unsigned short)fgetc(fp) + (((unsigned short)fgetc(fp))<<8);
      chunk_length = (unsigned long)fgetc(fp) + (((unsigned long)fgetc(fp))<<8) +
                     (((unsigned long)fgetc(fp))<<16) + (((unsigned long)fgetc(fp))<<24);
      switch(chunk_id)
      {
          case  0x4d4d : /* main chunk */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif        
                         /* set length of main chunk */
                         main_length = chunk_length;
			 break;
          case  0x3d3d : /* edit chunk */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif
			 break;        
          case  0x4000 : /*  edit object chunk */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif
                         /*  first element is a objectname terminated with \0 */
                         object_name[0] = '\0';
			 while(1)
			 {			 
                            letter = fgetc(fp);
			    if(letter == 0) break;
                            sprintf(object_name,"%s%c",object_name,letter);
			 }	 
                         printf("reading object %s ... \n",object_name);        
                         /* allocate memory for layer */
                         /* is first ? */
			 if( first_layer == NULL)
			 {
			    if((layer = (struct _object*) malloc(sizeof(struct _object)))==NULL)
			    {
			        error("not enough memory");
			    } 
			    first_layer = layer; 
                            layer->next = NULL;
			 }  
                         else
			 {
			    last_layer = layer;
			    if((layer = (struct _object*) malloc(sizeof(struct _object)))==NULL)
			    {
			        error("not enough memory");
			    } 
                            last_layer->next = layer;
			 }  
                         /* storing object name to struct */
                         sprintf(layer->name,"%s",object_name);
                         layer->type = 0; /* layer unknown */ 
			 break;        
          case  0x4100 : /*  edit object trimesh chunk */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif
                         layer->type = 1; /* set type mesh */
			 break;        
          case  0x4110 : /*  edit object trimesh vertices list chunk */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif

                         /* how many vertices ? */
                         layer->vertices = (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);
                         /* allocate memory for verices */ 
                         if((layer->x = (float*)malloc(sizeof(float)*layer->vertices))==NULL)
			 {
			     error("not enough memory");
			 } 
                         if((layer->y = (float*)malloc(sizeof(float)*layer->vertices))==NULL)
			 {
			     error("not enough memory");
			 } 
                         if((layer->z = (float*)malloc(sizeof(float)*layer->vertices))==NULL)
			 {
			     error("not enough memory");
			 } 
                         /* read vertices */
                         for(x=0;x<layer->vertices;x++)
                         {
			     fread(&(layer->x[x]),sizeof(float),1,fp);  
			     fread(&(layer->y[x]),sizeof(float),1,fp);  
			     fread(&(layer->z[x]),sizeof(float),1,fp);  			      			     
			 }   
                         #ifdef DEBUG
                         printf("number of vertices : %d\n",layer->vertices);
			 #endif
        		 break;        
          case 0x4120  : /*  edit object trimesh face list chunk */
                         #ifdef DEBUG
                         printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif
	        	
                         /* how many faces ? */
                         layer->faces = (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);			 
                         /* allocate memory fofr faces */
                         if((layer->face = (struct _face*)malloc(sizeof(struct _face)*layer->faces))==NULL)
			 {
			     error("not enough memory");
			 } 
                         /* read faces */
                         for(x=0;x<layer->faces;x++)
                         {
                            layer->face[x].a =  (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);
                            layer->face[x].b =  (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);
                            layer->face[x].c =  (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);
                            layer->face[x].options = (unsigned int)fgetc(fp) + (((unsigned int)fgetc(fp))<<8);
			 }   
			  
                         #ifdef DEBUG
                         printf("number of faces : %d\n",layer->faces);
	         	 #endif
			 break;        
          case 0x4600  : /*  edit object light */
                         #ifdef DEBUG
                         printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
			 #endif
                         layer->type = 2; /* set lamp */ 	        	
                         /* allocate x,y,z for lamp and spot */
			 if((layer->x = (float*)malloc(sizeof(float)*2))==NULL)
			 {
			     error("not enough memory");
			 } 
                         if((layer->y = (float*)malloc(sizeof(float)*2))==NULL)
			 {
			     error("not enough memory");
			 } 
                         if((layer->z = (float*)malloc(sizeof(float)*2))==NULL)
			 {
			     error("not enough memory");
			 }
                         /* read light position */
                         fread(&(layer->x[0]),sizeof(float),1,fp);
    			 fread(&(layer->y[0]),sizeof(float),1,fp);
			 fread(&(layer->z[0]),sizeof(float),1,fp);
		  
			 break;        
          case 0x4610  : /*  edit object light */
                         #ifdef DEBUG
                         printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
              		 #endif
                         /* check fo light */
			 if(layer->type != 2)
			 {
			    error("hmm ...  file seems to be corrupt");
			 }
                         layer->type = 3; /* set lamp */ 	        	
                         /* allocate x,y,z for lamp and spot */
                         /* read spot position */
                         fread(&(layer->x[1]),sizeof(float),1,fp);
    			 fread(&(layer->y[1]),sizeof(float),1,fp);
			 fread(&(layer->z[1]),sizeof(float),1,fp);
                         /* read hotspot */
                         fread(&(layer->hotspot),sizeof(float),1,fp); 
                         /* read falloff */
			 fread(&(layer->falloff),sizeof(float),1,fp);
		  
                  	 break;        
          default      : /* all the others */
                         #ifdef DEBUG
	                 printf("found chunkid %X at %X with length %X\n",
			         chunk_id,ftell(fp)-6,chunk_length);
          		 #endif        
	  		 fseek (fp,(long)(chunk_length - 6),1);  

      }		     
   }
   fclose(fp);
   
   /* write to vidioscape */
   layer = first_layer;
   n = 1; /* number of output files */
   while(layer != NULL)
   {
      /* data in layer ? */
      if(layer->type == 1 ) /* in this version meshes only */
      {
         sprintf(out_file,"%s.obj%d",out_name,n);      

         if((fp_out = fopen(out_file,"w"))==NULL)
         {
            error("can't create outputfile");
         } 
         printf("writing object %s to %s ...\n",layer->name,out_file);
         switch(layer->type)
	 {
             case 1 :  /* write Vidioscape mesh file */      
                       fprintf(fp_out,"3DG1\n");
                       fprintf(fp_out,"%d\n",layer->vertices);
                       /* write vertices */
                       for(x=0;x<layer->vertices;x++)
                       {
                          fprintf(fp_out,"%f %f %f\n",layer->x[x],layer->y[x],layer->z[x]);
                       }	   
                       /* write faces */
	               for(x=0;x<layer->faces;x++)
                       {
                          fprintf(fp_out,"3 %d %d %d 0xcccccc\n",layer->face[x].a,layer->face[x].b,layer->face[x].c);            
                       }
                       break;

	    case 2 :   /* write Vidioscape lamp file */
                       /* later */   	              
		       break;

            case 3 :   /* write Vidioscape spot file */
                       /* later */      
		       break;		      

         } /* end switch */
         fclose(fp_out);
         n++; 
      } 	 
      layer = layer->next;
   }    
   printf("done.\n");
}
